"""Frontmatter module."""
import re

from ruamel.yaml import YAML

MARKUPS = {
    "YAML": re.compile(rb"(?:^---\s*\n)([\s\S]+?)(?:^(?:---|\.\.\.)\s*\n)", re.MULTILINE)}

PARSERS = {"YAML": YAML().load}


def parse(body):
    text = body.strip(b"\n")
    lang = None

    lang = select_language(text)
    if lang is None:
        return {}, text

    __, frontmatter, body = MARKUPS[lang].split(text, 2)
    frontmatter = frontmatter.strip(b"\n")
    body = body.strip(b"\n")

    frontmatter = PARSERS[lang](frontmatter)

    return frontmatter, body


def select_language(text):
    for lang, regex in MARKUPS.items():
        if regex.match(text):
            return lang



"""
:mod:`taika.ext.rst` -- ReStructuredText
========================================

This extension parses the content of the documents into HTML using ReStructuredText specifications.

Trigger
-------

This extension is subscribed to the "doc-post-read" event.

Frontmatter
-----------

None.

Process
-------

#. Reads the suffix of `path` and if it matches, process the document.
#. Modifies the suffix of `url` path to ".html".
#. Process the content with :func:`docutils.publish_parts` and replaces it with the "body" part.
#. Done!

Configuration
-------------

.. data:: rst_suffix (list)

    Default: **[.rst]**

    Tells the parser to ONLY modify docs with that suffix. Otherwise the document is ignored.
    This is checked against the source path (`path`), not the destination path (`url`).

.. data:: rst_strict (bool)

    Default: **True**

    Exits with error code 1 if there is any warning or error when parsing files.


.. data:: rst_options (dict)

    Default:
        | stylesheet_path: ''
        | halt_level: 1
        | traceback: True
        | report_level: 5
        | syntax_highlight: 'short'
        | doctitle_xform: False

    You can check the available options at `HTML writer documentation
    <http://docutils.sourceforge.net/docs/user/config.html#html4css1-writer>`_


Functions
---------
"""
import logging
import sys
from io import StringIO

try:
    from docutils.core import publish_parts
    from docutils.utils import SystemMessage
except ImportError:
    print("'taika.ext.rst' needs the package 'docutils' to be installed.")
    sys.exit(1)


RST_OPTIONS = {
    "stylesheet_path": "",
    "halt_level": 1,
    "traceback": True,
    "report_level": 5,
    "syntax_highlight": "short",
    "doctitle_xform": False,
}
DEFAULT_SUFFIX = ".rst"
OUT_SUFFIX = ".html"
RST_STRICT = True

LOGGER = logging.getLogger(__name__)


def parse_rst(site, document):
    """Parse ``content`` and modify ``url`` keys of `document`.

    Parameters
    ----------
    site : :class:`taika.taika.Taika`
        The Taika site.
    document : dict
        The document to be parsed.
    """
    suffixes = site.config.get("rst_suffix", DEFAULT_SUFFIX)
    rst_options = site.config.get("rst_options", RST_OPTIONS)

    if document["path"].suffix not in suffixes:
        return

    document["url"] = document["url"].with_suffix(OUT_SUFFIX)
    warning_stream = StringIO()
    rst_options["warning_stream"] = warning_stream
    try:
        html = publish_parts(
            source=document["content"],
            source_path=site.source / document["path"],
            writer_name="html5",
            settings_overrides=rst_options,
        )["body"]

        document["content"] = html
    except SystemMessage:
        msg = warning_stream.getvalue().strip("\n").replace("\n", " ")
        rst_strict = site.config.get("rst_strict", RST_STRICT)
        if rst_strict:
            action = "Exiting"
            LOGGER.critical(f"Problem when parsing a file. Error message: '{msg}' {action}...")
            sys.exit(1)
        else:
            action = f"Skipping, 'rst_strict' = {rst_strict}"
            LOGGER.warning(f"Problem when parsing a file. Error message: '{msg}' {action}...")


def setup(site):
    site.events.register("doc-post-read", parse_rst)



r"""
:mod:`taika.ext.layouts` -- Jinja layouts
=========================================

This extension renders documents content trought a the Jinja2 templating engine. It also
renders the content of documents itself if any Jinja block/comment/var is detected.

Event
-----

This extension is subscribed to the :data:`site-post-read` event.

Payload
-------

When the content and the templates are rendered, certain payload is passed and becomes
accessible by both content and templates. This payload has two main keys: ``site`` and
``document``.

Using :code:`document` you can access the document attributes being processed, such as the
:code:`path`, :code:`content`, etc. Check :doc:`/reference/documents` for details.

Inside ``site``, the :class:`taika.Taika` is accessible.

.. note::

    Note that ``site.config`` returns a dictionary with all the sections included, so to access
    which extensions are listed you should use ``site.config.taika.extensions``. This is a long
    "import-like" statement, and probably we will shrink it in the future.

Frontmatter
-----------

.. data:: layout (str)

    The layout that should render the document. Should exist under the :data:`layouts_path`. If None
    the documents is not passed throught the template, but its body is still rendered.

Configuration
-------------

.. data:: layouts_path (dangling-list)

    Default: **./templates/**

    A list of paths from where the layouts will be loaded.

.. data:: layouts_options (dict)

    Default: **{}** (empty-dict)

    A dictionary (ini-style) of options to pass to the Jinja environment when created.

.. data:: layouts_default (str)

    Default: **index.html**

    The default layout if the document has no :code:`layout` defined.

.. data:: layouts_pattern (str)

    Default: *****

    Which files should be renderered. Default to all the files.

Default filters
---------------

Jinja

Process
-------

#. (pre-registering) The :class:`JinjaRenderer` is initialized with the configuration. The Jinja
   environment is created and the templates loaded.
#. Checks if the path of the document matches :data:`layouts_pattern`, if not, skips it.
#. Composes the layout using the document itself, so the document metadata is available directly.
#. If the content has any Jinja flag, it is renderered, so you can include Jinja syntax into the
   document text.
#. Then the content (rendered or not) is rendered throught the template :data:`layout`.
#. The document's content is modified.
#. Done!

Classes and Functions
---------------------
"""
import fnmatch
import logging
import platform
import sys
from pathlib import Path

try:
    import jinja2
except ImportError:
    print("This extension needs Jinja2, please install the jinja2 module.")
    sys.exit(1)

LAYOUTS_PATH = "./templates/"
LAYOUTS_DEFAULT = "index.html"
LAYOUTS_OPTIONS = {}
LAYOUTS_PATTERN = "*"

LOGGER = logging.getLogger(__name__)


class DocumentNotFound(Exception):
    pass


class JinjaRenderer(object):
    """This class holds the Jinja2 environment, removing the need to create it each time.

    Attributes
    ----------
    env : :class:`jinja2.Environment`
        The configured Jinja environment.
    layouts_pattern : str
        The pattern that the `path` attribute of documents should match in order to be processed.
    layouts_default : str
        The option so the :meth:`JinjaRenderer.render_content` can access it.
    """

    def __init__(self, config):
        layouts_path = config.get("layouts_path", LAYOUTS_PATH)
        layouts_options = config.get("layouts_options", LAYOUTS_OPTIONS)

        loader = jinja2.FileSystemLoader(layouts_path)
        self.env = jinja2.Environment(loader=loader, **layouts_options)

        self.layouts_default = config.get("layouts_default", LAYOUTS_DEFAULT)
        self.layouts_pattern = config.get("layouts_pattern", LAYOUTS_PATTERN)

    def render_content(self, site):
        payload = {}
        payload["site"] = site

        for document in site.documents:
            if not fnmatch.fnmatch(document["path"], self.layouts_pattern):
                continue

            payload["document"] = document
            content = document["content"]

            if self._has_block(content) or self._has_comment(content) or self._has_var(content):
                content = self.env.from_string(content).render(**payload)

            document["pre_render_content"] = payload["document"]["content"] = content

            template_name = document.get("layout", self.layouts_default)
            if template_name is not None:
                template = self.env.get_template(template_name)
                content = template.render(**payload)

            document["content"] = content

    def _has_block(self, s):
        return self.env.block_start_string in s and self.env.block_end_string in s

    def _has_var(self, s):
        return self.env.variable_start_string in s and self.env.variable_end_string in s

    def _has_comment(self, s):
        return self.env.comment_start_string in s and self.env.comment_end_string in s


@jinja2.contextfilter
def link(context, path):
    is_absolute = False

    path = Path(path)
    if platform.system() == "Windows":
        path = path.resolve()

    if path.is_absolute():
        path = Path(*path.parts[1:])
        is_absolute = True

    LOGGER.debug(
        f"Link ContextFilter --> From:'{context['document']['path']}' the document '{path}' is"
        "requested."
    )
    for doc in context["site"].documents:
        if doc["path"] == path:
            return_path = doc["url"]
            break
    else:
        raise DocumentNotFound(path)

    if is_absolute:
        return "/" + str(return_path)
    return return_path


def setup(site):
    renderer = JinjaRenderer(site.config)
    renderer.env.filters.update({"link": link})
    site.events.register("site-post-read", renderer.render_content)



# -*- coding: utf-8 -*-

"""
:mod:`taika`
============

The top-level package contain some meta info about the package to be accessible by other tools.

.. data:: __author__ (str)

    The author name.

.. data:: __email__ (str)

    The email of `__author__`.

.. data:: __version__ (str)

    The version of the package.
"""

from taika.taika import Taika

__all__ = ["__author__", "__email__", "__version__"]
__all__ += ["Taika"]


__author__ = """Hector Martinez-Lopez"""
__email__ = "hector.martinez.ub@gmail.com"
__version__ = "0.6.0"



# -*- encoding: utf-8 -*-

"""
:mod:`taika.taika`
==================
"""
import logging
import sys
from importlib import import_module
from pathlib import Path

import ruamel.yaml as yaml

from taika import frontmatter
from taika.events import EventManager
from taika.utils import add_syspath
from taika.utils import pretty_json

__all__ = ["read_conf", "write_file", "read_file"]

logging.basicConfig(level=logging.DEBUG)
LOGGER = logging.getLogger("taika")

TAIKA_CONF = "{source}/taika.yml"

DEFAULT_EXTENSIONS = ""
DEFAULT_EXTENSIONS_PATH = ""


class Taika(object):
    """Taika main class.

    Attributes
    ----------
    source : :class:`pathlib.Path`
    destination : :class:`pathlib.Path`
    events : :class:`taika.events.EventManager`
    config : dict
    documents : list
    """

    def __init__(self, source, destination, conf_path=None):
        """Read all RST files from `source`, parse them and write them back as HTML in `dest`.

        Parameters
        ----------
        source : str
            The path to the source directory.
        dest : str
            The path where the parsed files will be writed.
        conf_path : str, optional (default=None)
            The path where the configuration file lives. If :code:`None`, the default path
            will be used.

        Notes
        -----
        By default, this functions maintains the `source` structure in `dest`.
        """
        self.documents = []
        self.events = EventManager()

        if conf_path is None:
            conf_path = TAIKA_CONF

        conf_path = conf_path.replace("{source}", str(source))
        self.config = read_conf(conf_path)

        self.import_extensions()
        self.source = Path(source)
        self.destination = Path(destination)

    def import_extensions(self):
        """Load the configuration and extensions."""
        extension_paths = self.config.get("extension_paths", DEFAULT_EXTENSIONS_PATH)
        extensions = self.config.get("extensions", DEFAULT_EXTENSIONS)

        LOGGER.debug(f"extension_paths: {extension_paths}")
        LOGGER.debug(f"extensions: {extensions}")

        # Don't create byte-code, we want the extensions folders be clean.
        sys.dont_write_bytecode = True
        with add_syspath(extension_paths):
            for ext in extensions:
                import_module(ext).setup(self)
        sys.dont_write_bytecode = False

    def process(self):
        """Run :meth:`Taika.read` and :meth:`Taika.write`."""
        self.documents = self.read(self.source)
        self.events.call("site-post-read", self)
        self.write(self.documents, self.destination)

    def read(self, source):
        """Read all the files *recursively* from a `source` directory and load them as dictionaries.

        Parameters
        ----------
        source : :class:`pathlib.Path`
            The source directory where the documents are read from.

        Returns
        -------
        documents : list
            A list of dictionaries that represent documents.
        """
        documents = []
        for path in source.glob("**/*"):
            if path.is_dir():
                continue

            document = read_file(path)
            self.events.call("doc-post-read", self, document)
            documents.append(document)
        return documents

    def write(self, documents, destination):
        """Call `taika.taika.write_file` for each document on `documents` with `destination`.

        Parameters
        ----------
        documents : list
            A list of dictionaries that represent documents.
        destination : str or :class:`pathlib.Path`
            The destination directory.
        """
        for document in documents:
            write_file(document, destination)


def read_conf(conf_path):
    """Read the configuration file `conf_path`. It should be an INI style configuration.

    Parameters
    ----------
    conf_path : str
        The path to the configuration file to be readed.

    Returns
    -------
    conf : `configparser.ConfigParser`
        An instance of a ConfigParser which holds the configuration.

    Raises
    ------
    SystemExit
        If `conf_path` is not a file.
    """
    conf_path = Path(conf_path)
    if not conf_path.is_file():
        LOGGER.critical(f"The configuration file {conf_path} is not a file. Exiting...")
        exit(1)
    else:
        with open(conf_path) as fd:
            conf = yaml.safe_load(fd)

    LOGGER.debug(pretty_json(conf))

    return conf


def read_file(path):
    """Read `path` and return the document as a dictionary.

    Parameters
    ----------
    path : str or :class:`pathlib.Path`
        A path to a file to be read.

    Returns
    -------
    document : dict
        A dictionary that holds the information of the document read from `path`.
    """
    path = Path(path)

    LOGGER.debug("Reading: %s", path)

    with open(path, "rb") as fd:
        raw_content = fd.read()
    metadata, content = frontmatter.parse(raw_content)
    document = {
        "path": path.relative_to(path.parts[0]),  # the first part is the source directory
        "url": path.relative_to(path.parts[0]),
        "content": content,
        "raw_content": raw_content,
    }

    document.update(metadata)

    return document


def write_file(document, destination):
    """Given a `document` and a destionation, write `document.content` in the destination.

    Parameters
    ----------
    document : dict
        A dictionary representing a document. Should have ``content`` and ``url``.
    destination : str or :class:`pathlib.Path`
        The destination directory where the document will be written.

    Raises
    ------
    KeyError
        If the document doesn't have ``content`` or ``url``.
    """
    destination = Path(destination)

    path = document["url"]
    content = document["content"]

    path = destination.joinpath(path)

    path.parent.mkdir(parents=True, exist_ok=True)
    path.touch()
    try:
        with open(path, "wb") as fd:
            fd.write(content)
        LOGGER.debug("Writing: '%s' as bytes.", path)
    except TypeError:
        with open(path, "w", encoding="utf-8") as fd:
            fd.write(content)
        LOGGER.debug("Writing: '%s' as string.", path)



"""
:mod:`taika.ext.excerpt` -- Documents excerpts
==============================================

This extensions creates a excerpt for the documents based on it's content.

Event
-----

This extension is subscribed to the "doc-post-read" event.

Frontmatter
-----------

.. data:: excerpt_separator (str)

    Use this separator instead of the global separator defined in the configuration.

Configuration
-------------

.. code-block:: yaml

    excerpt_separator: <!-- read-more -->


.. data:: excerpt_separator (str)

    Default: :code:`None`

    A string that will be used as excerpt separator. Default to :code:`None` so no excerpt
    will be generated.


Process
-------

#. Check for the frontmatter option, otherwise use the global or the default separator.
#. If separator is None, the first :code:`<p>` tag is retrieved if existent.
#. If the first :code:`<p>` tag is not found, :code:`\\n\\n` (double line separator) is
   used as separator.
#. If separator is something, the text before that separator is retrieved if existent.
#. The :code:`excerpt` is inserted into the document so it will be accessible.

Classes and Functions
---------------------
"""
import logging
import sys

try:
    import bs4
except ImportError:
    print("taika.ext.excerpt needs beautifulsoup4 to be installed.")
    sys.exit(1)

DEFAULT_SEPARATOR = None
LOGGER = logging.getLogger(__name__)


def get_excerpt(site, document):
    separator = site.config.get("excerpt_separator", DEFAULT_SEPARATOR)
    if document["url"].suffix != ".html":
        LOGGER.debug(f"Document {document['url']} doesn't finish in html.")
        return

    if "excerpt_separator" in document:
        separator = document["excerpt_separator"]

    if separator is None:
        html = bs4.BeautifulSoup(document["content"], features="html.parser")
        first_p = html.find("p")

        if first_p is not None:
            document["excerpt"] = "<p>" + first_p.get_text() + "</p>"
            LOGGER.debug(f"Extracting the first paragraph for {document['url']}.")
        else:
            LOGGER.debug(f" Didn't found <p> tag. Using \\n\\n as separator.")
            document["excerpt"] = _text_before("\n\n", document["content"])
    else:
        LOGGER.debug(f"Using separator '{separator}' for document {document['url']}.")
        document["excerpt"] = _text_before(separator, document["content"])


def _text_before(string, text):
    pos = text.find(string)
    if pos == -1:
        LOGGER.debug(f"Separator {string} not found on text.")
        LOGGER.debug(text)

    return text[:pos]


def setup(site):
    site.events.register("doc-post-read", get_excerpt)



import contextlib
import json
import sys
from functools import partial

__all__ = ["add_syspath", "pretty_json"]

pretty_json = partial(json.dumps, indent=4)


@contextlib.contextmanager
def add_syspath(paths):
    for path in paths:
        sys.path.insert(0, path)

    yield

    for path in paths:
        sys.path.remove(path)



"""
:mod:`taika.ext.collections` -- Grouping content
================================================

This extension groups documents using patterns specified by the user. It also order those
documents using certain keys specified by the user.

It uses the patterns listed from top to bottom, the documents not included in the first
pattern are not matched against the second pattern, so be liberal in the first pattern
and more restricted at the bottom. Also, if the pattern starts with ``!`` (exclamation mark)
the documents matching will be excluded. For example:

.. code-block:: yaml

    collections:
    posts:
      patterns:
        - "posts/*" # Include all under posts
        - "!posts/index.rst" # Ignore posts/index.rst

Event
-----

This extension is subscribed to the :data:`site-post-read` event.

Process
-------

#. Setup where the ``collections`` keys is retrieved.
#. When the extension is called, scans the documents checking their ``path``.
#. If ``path`` matches the patterns provided, it's added to the collection.
#. Finally, the attribute ``collections`` is created on :class:`taika.Taika`.

Configuration
-------------

.. code-block:: yaml

    # Match all but the index.rst file on posts/
    collections:
      posts:
        patterns:
          - "posts/*" # Include all under posts
          - "!posts/index.rst" # Ignore posts/index.rst

.. data:: collections (dict)

    Default: **{}**

    A dictionary where each key specifies the name of the collection.

.. data:: collection.patterns (list)

    Default: [''] (empty string)

    The patterns to be used in order to group the files. By default, it matches nothing.

Classes and Functions
---------------------
"""
import fnmatch
import logging
from collections import defaultdict

COLLECTIONS_DEFAULT = {}
DEFAULT_PATTERNS = [""]

LOGGER = logging.getLogger(__name__)


def match(path, patterns, reverse_character="!"):
    for pattern in patterns:
        match = fnmatch.fnmatch(path, pattern.lstrip(reverse_character))
        if pattern.startswith(reverse_character):
            match = not match

        if not match:
            return False

    return True


class Collector(object):
    """Main class which retrieves the configuration and the organize the documents."""

    def __init__(self, config):
        self.collections_config = config.get("collections", COLLECTIONS_DEFAULT)

    def organize(self, site):
        """Classify the documents and creates the collections attribute on `site`."""
        collections = _organize(site.documents, self.collections_config)
        site.collections = collections


def _organize(documents, config):
    collections = defaultdict(list)
    for document in documents:
        for collection_name, collection_options in config.items():
            patterns = collection_options.get("patterns", DEFAULT_PATTERNS)
            included = match(document["path"], patterns)
            if included:
                LOGGER.debug(f"Document '{document['path']}' --> Collection '{collection_name}'.")
                collections[collection_name].append(document)

    return collections


def setup(site):
    collector = Collector(site.config)
    site.events.register("site-post-read", collector.organize)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['taika', 'taika.ext']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4', 'docutils', 'jinja2', 'ruamel.yaml']

entry_points = \
{'console_scripts': ['taika = taika.cli:main']}

setup_kwargs = {
    'name': 'taika',
    'version': '0.6.0',
    'description': 'Another Static Site Generator',
    'long_description': 'Taika\n=====\n\n.. image:: https://img.shields.io/pypi/v/taika.svg\n    :target: https://pypi.python.org/pypi/taika\n\n.. image:: https://readthedocs.org/projects/taika/badge/?version=latest\n    :target: https://taika.readthedocs.io/en/latest/?badge=latest\n    :alt: Documentation Status\n\n.. image:: https://gitlab.com/hectormartinez/taika/badges/master/pipeline.svg\n    :target: https://gitlab.com/hectormartinez/taika/commits/master\n    :alt: Pipeline Status\n\n.. image:: https://gitlab.com/hectormartinez/taika/badges/master/coverage.svg\n    :target: https://gitlab.com/hectormartinez/taika/commits/master\n    :alt: Coverage Report\n\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/ambv/black\n   :alt: Code style: black\n\n\nTaika in another Static Site Generator which I created as a learning activity.\nToday I use it to build my personal page.\n\nFeatures\n--------\n\n* YAML configuration\n* Jinja2 templates\n\nLicense\n-------\n\n`MIT <./LICENSE>`__\n\n',
    'author': 'Hector Martinez',
    'author_email': 'hector.martinez.ub@gmail.com',
    'url': 'https://gitlab.com/hectormartinez/taika',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-

"""
:mod:`taika.cli`
================

The entry point for the command line interface of Taika.
"""
import argparse
import os
import sys

from taika import Taika
from taika.taika import TAIKA_CONF


def parse_arguments(arguments):
    """Create a :class:`argparse.ArgumentParser` and run :meth:`argparse.ArgumentParser.parse_args`
    agains `arguments`.

    Parameters
    ----------
    arguments : list
        A list of arguments to be parsed.

    Returns
    -------
    namespace : `argparse.Namespace`
        The namespace created when `arguments` are parsed.
    """
    parser = argparse.ArgumentParser(prog="taika")
    parser.add_argument("source", help="The directory that contains the files to be parsed.")
    parser.add_argument("destination", help="The directory that will contain the parsed files.")
    parser.add_argument(
        "-c",
        "--conf",
        default=os.getenv("TAIKA_CONF"),
        help=f"The configuration file for Taika. Default to '{TAIKA_CONF}', but it can be "
        "setted using the environment variable 'TAIKA_CONF'.",
    )

    return parser.parse_args(arguments)


def main(arguments=None):
    """The main entry point, parse `arguments` behaves accordingly.

    Parameters
    ----------
    arguments : list, optional (default=None)
        A list of arguments to be parsed. If ``None``, ``sys.argv[1:]`` is used.

    Returns
    -------
    err_code : int
        Non-zero value indicates error, or zero on success.
    """
    if arguments is None:
        arguments = sys.argv[1:]

    args = parse_arguments(arguments)
    site = Taika(args.source, args.destination, args.conf)
    site.process()
    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



"""
:mod:`taika.events` -- Basic event managment
============================================

This module offers a simple event manager implemented in the class
:class:`taika.events.EventManager`.
"""
from collections import defaultdict

__all__ = ["events", "EventNotFound", "EventManager"]

events = set(["doc-post-read", "site-post-read"])
"""The events that the event manager can register functions to."""


class EventNotFound(Exception):
    """Exception raised when an event does not exists."""


class EventManager(object):
    """Register functions to events and passes them arguments and keyword arguments when called."""

    def __init__(self):
        self.funcs = defaultdict(dict)
        self.next_id = 0

    def register(self, event, func):
        """Register a callable `func` to an `event`.

        Parameters
        ----------
        event : str
            The event to which `func` will be registered.
        func : callable
            A callable that will recieve arguments and keywords arguments when `event` is triggered.

        Returns
        -------
        current_id : int
            The ID assigned to the function.

        Raises
        ------
        :exc:`EventNotFound`
            If the `event` that is being triggered does not exist.
        """
        _event_exists(event)

        current_id = self.next_id
        self.next_id += 1
        self.funcs[event][current_id] = func
        return current_id

    def call(self, event, *args, **kwargs):
        r"""Call all the functions registered to `event` passing `*args` and `*kwargs`.

        Parameters
        ----------
        event : str
            The event which be triggered.

        Raises
        ------
        :exc:`EventNotFound`
            If the `event` that is being triggered does not exist.
        """
        _event_exists(event)

        results = []
        for func in self.funcs[event].values():
            results.append(func(*args, **kwargs))
        return results


def _event_exists(event):
    """Check if `event` is inside :data:`events`.

    Parameters
    ----------
    event : str
        The name of the event.

    Raises
    ------
    :exc:`EventNotFound`
        When the event is not in :data:`events`.
    """
    if event not in events:
        raise EventNotFound(f"The event '{event}' does not exists.'")
    return True
